/*
 * sox_reader.c
 * Copyright (C) 2010 - 2013 Peter Belkner <pbelkner@snafu.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "ffsox.h"

#if 0 // { libav
int av_frame_get_channels(const AVFrame *frame)
{
  return av_get_channel_layout_nb_channels(frame->channel_layout);
}
#endif // }

///////////////////////////////////////////////////////////////////////////////
typedef uint8_t u8_t;
typedef int16_t s16_t;
typedef int32_t s32_t;
typedef float f32_t;
typedef double f64_t;

typedef struct priv {
#if defined (FFSOX_LOCAL_AR) // {
  ffsox_ar_t ar;
#else // } {
  ffsox_ar_t *ar;
#endif // }

#if 0 // [
#define PRIV_UNION
  union {
    // interleaved
    struct {
      uint8_t *rp;
      uint8_t *mrp;
    } i;

    // planar
    struct {
      uint8_t **rp;
      size_t nch;
      size_t bytes;
      size_t n;
      size_t i;
    } p;
  };
#else // ] [
  AVFrame *frame;
  int sample;
  int channel;
#endif // ]
} priv_t;

///////////////////////////////////////////////////////////////////////////////
#if defined (PRIV_UNION) // [
static void priv_reset_i(priv_t *priv)
{
#if defined (FFSOX_LOCAL_AR) // {
  AVFrame *frame=priv->ar.frame;
#else // } {
  AVFrame *frame=ffsox_ar_frame_internal(priv->ar);
#endif // }
  size_t n=frame->nb_samples;
#if defined (FFSOX_DEPRECATED) // [
  size_t nch=frame->channels;
#else // ] [
  size_t nch=av_frame_get_channels(frame);
#endif // ]
  size_t bytes=av_get_bytes_per_sample(frame->format);
  uint8_t *rp=frame->data[0];

DVWRITELN("n=%u nch=%u bytes=%u",n,nch,bytes);
  priv->i.rp=rp;
  priv->i.mrp=rp+n*nch*bytes;
}
#endif // ]

#if defined (PRIV_UNION) // [
static void priv_reset_p(priv_t *priv)
{
#if defined (FFSOX_LOCAL_AR) // {
  AVFrame *frame=priv->ar.frame;
#else // } {
  AVFrame *frame=ffsox_ar_frame_internal(priv->ar);
#endif // }

  priv->p.rp=frame->data;
#if defined (FFSOX_DEPRECATED) // [
  priv->p.nch=frame->channels;
#else // ] [
  priv->p.nch=av_frame_get_channels(frame);
#endif // ]
  priv->p.bytes=av_get_bytes_per_sample(frame->format);
  priv->p.n=frame->nb_samples;
  priv->p.i=0;
}
#endif // ]

#if 0 // [
#define PRIV_RESET
static void priv_reset(priv_t *priv)
{
#if defined (FFSOX_LOCAL_AR) // {
  AVFrame *frame=priv->ar.frame;
#else // } {
  AVFrame *frame=ffsox_ar_frame_internal(priv->ar);
#endif // }
  int sample_fmt;

  if (!frame) {
    DMESSAGE("frame is NULL");
    return;
  }

  sample_fmt=frame->format;

  switch (sample_fmt) {
  // interleaved:
  case AV_SAMPLE_FMT_U8:
  case AV_SAMPLE_FMT_S16:
  case AV_SAMPLE_FMT_S32:
  case AV_SAMPLE_FMT_FLT:
  case AV_SAMPLE_FMT_DBL:
    priv_reset_i(priv);
    break;
  // planar:
  case AV_SAMPLE_FMT_U8P:
  case AV_SAMPLE_FMT_S16P:
  case AV_SAMPLE_FMT_S32P:
  case AV_SAMPLE_FMT_FLTP:
  case AV_SAMPLE_FMT_DBLP:
    priv_reset_p(priv);
    break;
  // unsupported:
  default:
    DVWRITELN("unsupported sample format: %s",
        av_get_sample_fmt_name(sample_fmt));
    break;
  }
}
#endif // ]

///////////////////////////////////////////////////////////////////////////////
static int priv_getopts(sox_effect_t *effp, int argc, char *argv[])
{
  priv_t *priv=effp->priv;
#if defined (FFSOX_LOCAL_AR) // {
  ffi_t *ffi;
#endif // }
  int n=1;

DMARKLN();
  memset(priv,0,sizeof *priv);

  if (argc<2)
    return SOX_EOF;

#if defined (FFSOX_LOCAL_AR) // {
  ffi=(ffi_t *)argv[n++];
#else // } {
  priv->ar=(ffsox_ar_t *)argv[n++];
#endif // }

#if defined (FFSOX_LOCAL_AR) // {
  if (NULL==ffsox_ar_init_ffi(&priv->ar,ffi))
    return SOX_EOF;
#endif // }

#if defined (PRIV_RESET) // [
  priv_reset(priv);
#endif // ]

  return SOX_SUCCESS;
}

#if defined (FFSOX_LOCAL_AR) // {
static int priv_kill(sox_effect_t *effp)
{
  priv_t *priv=effp->priv;

  ffsox_ar_cleanup(&priv->ar);

  return SOX_SUCCESS;
}
#endif // }

///////////////////////////////////////////////////////////////////////////////
#if defined (FFSOX_LOCAL_AR) // {
#define priv_ar(priv) \
  (&(priv)->ar)
#else // } {
#define priv_ar(priv) \
  ((priv)->ar)
#endif // }
#define priv_scale_s16(q) \
 (-(q)/INT16_MIN)
#define priv_scale_s32(q) \
 (-(q)/INT32_MIN)
#define priv_scale_f32(q) \
 (q)
#define priv_scale_f64(q) \
 (q)

///////////////////////////////////////////////////////////////////////////////
#if defined (PRIV_UNION) // [
#if defined (LIBFFSOX_BS1770) // [
#define PRIV_DRAIN_I(id1,id2,ID) \
static int priv_drain_##id1##i(sox_effect_t *effp, int nch, \
    sox_sample_t *obuf, size_t *osamp) \
{ \
  priv_t *priv=effp->priv; \
  ffsox_ar_t *ar=priv_ar(priv); \
  double q1=ar->q; \
  double q2=priv_scale_##id1(q1); \
  uint8_t *rp,*mrp; \
  sox_sample_t *wp=obuf; \
  sox_sample_t *mwp=obuf+(*osamp-=*osamp%nch); \
  int ch; \
  id1##_t x; \
  double y; \
  bs1770_sample_##id2##_t sample; \
  SOX_SAMPLE_LOCALS; \
 \
  if (0==ar->count) { \
    priv->i.rp=NULL; \
    priv->i.mrp=NULL; \
  } \
 \
  rp=priv->i.rp; \
  mrp=priv->i.mrp; \
 \
  while (wp<mwp) { \
    for (ch=0;ch<nch;++ch) { \
      while (rp==mrp) { \
        ffsox_ar_next(ar); \
 \
        if (!ffsox_ar_valid(ar)) \
          goto done; \
 \
        priv_reset_i(priv); \
        rp=priv->i.rp; \
        mrp=priv->i.mrp; \
      } \
 \
      memcpy(&x,rp,sizeof x); \
      rp+=sizeof x; \
      sample[ch]=x; \
 \
      if (1.0==q1) \
        *wp++=SOX_##ID##_TO_SAMPLE(x,effp->clips); \
      else { \
        y=q2*x; \
        *wp++=SOX_FLOAT_64BIT_TO_SAMPLE(y,effp->clips); \
      } \
    } \
 \
    if (NULL!=ar->lufs) \
      ar->lufs->vmt->add_sample_##id2(ar->lufs,sample); \
  } \
done: \
 \
  priv->i.rp=rp; \
 \
  return 0<(*osamp=wp-obuf)?SOX_SUCCESS:SOX_EOF; \
}
#else // ] [
#define PRIV_DRAIN_I(id1,id2,ID) \
static int priv_drain_##id1##i(sox_effect_t *effp, int nch, \
    sox_sample_t *obuf, size_t *osamp) \
{ \
  priv_t *priv=effp->priv; \
  ffsox_ar_t *ar=priv_ar(priv); \
  double q1=ar->q; \
  double q2=priv_scale_##id1(q1); \
  uint8_t *rp,*mrp; \
  sox_sample_t *wp=obuf; \
  sox_sample_t *mwp=obuf+(*osamp-=*osamp%nch); \
  int ch; \
  id1##_t x; \
  double y; \
  bs1770_sample_##id2##_t sample; \
  SOX_SAMPLE_LOCALS; \
 \
  if (0==ar->count) { \
    priv->i.rp=NULL; \
    priv->i.mrp=NULL; \
  } \
 \
  rp=priv->i.rp; \
  mrp=priv->i.mrp; \
 \
  while (wp<mwp) { \
    for (ch=0;ch<nch;++ch) { \
      while (rp==mrp) { \
        ffsox_ar_next(ar); \
 \
        if (!ffsox_ar_valid(ar)) \
          goto done; \
 \
        priv_reset_i(priv); \
        rp=priv->i.rp; \
        mrp=priv->i.mrp; \
      } \
 \
      memcpy(&x,rp,sizeof x); \
      rp+=sizeof x; \
      sample[ch]=x; \
 \
      if (1.0==q1) \
        *wp++=SOX_##ID##_TO_SAMPLE(x,effp->clips); \
      else { \
        y=q2*x; \
        *wp++=SOX_FLOAT_64BIT_TO_SAMPLE(y,effp->clips); \
      } \
    } \
 \
    if (NULL!=ar->lufs) \
      ar->lufs->vmt->add_sample_##id2(ar->lufs,sample); \
  } \
done: \
 \
  priv->i.rp=rp; \
 \
  return 0<(*osamp=wp-obuf)?SOX_SUCCESS:SOX_EOF; \
}
#endif // ]

PRIV_DRAIN_I(s16,i16,SIGNED_16BIT)
PRIV_DRAIN_I(s32,i32,SIGNED_32BIT)
PRIV_DRAIN_I(f32,f32,FLOAT_32BIT)
PRIV_DRAIN_I(f64,f64,FLOAT_64BIT)
#else // ] [
#if defined (PRIV_UNION) // [
#define PRIV_DRAIN_I(id,ID) \
static int priv_drain_##id##i(sox_effect_t *effp, int nch, \
    sox_sample_t *obuf, size_t *osamp) \
{ \
  priv_t *priv=effp->priv; \
  ffsox_ar_t *ar=priv_ar(priv); \
  double q1=ar->q; \
  double q2=priv_scale_##id(q1); \
  uint8_t *rp,*mrp; \
  sox_sample_t *wp=obuf; \
  sox_sample_t *mwp=obuf+(*osamp-=*osamp%nch); \
  id##_t x; \
  double y; \
  SOX_SAMPLE_LOCALS; \
 \
  if (0==ar->count) { \
    priv->i.rp=NULL; \
    priv->i.mrp=NULL; \
  } \
 \
  rp=priv->i.rp; \
  mrp=priv->i.mrp; \
 \
  while (wp<mwp) { \
    while (rp==mrp) { \
      ffsox_ar_next(ar); \
 \
      if (!ffsox_ar_valid(ar)) \
        goto done; \
 \
      priv_reset_i(priv); \
      rp=priv->i.rp; \
      mrp=priv->i.mrp; \
    } \
 \
    memcpy(&x,rp,sizeof x); \
    rp+=sizeof x; \
 \
    if (1.0==q1) \
      *wp++=SOX_##ID##_TO_SAMPLE(x,effp->clips); \
    else { \
      y=q2*x; \
      *wp++=SOX_FLOAT_64BIT_TO_SAMPLE(y,effp->clips); \
    } \
  } \
done: \
  \
  priv->i.rp=rp; \
 \
  return 0<(*osamp=wp-obuf)?SOX_SUCCESS:SOX_EOF; \
}
#else // ] [
#define PRIV_DRAIN_I(id,ID) \
static int priv_drain_##id##i(sox_effect_t *effp, sox_sample_t *obuf, \
    size_t *osamp) \
{ \
  priv_t *priv=effp->priv; \
  AVFrame *frame=priv->frame; \
  ffsox_ar_t *ar=priv_ar(priv); \
  double q1=ar->q; \
  double q2=priv_scale_##id(q1); \
  int sample; \
  sox_sample_t *wp=obuf; \
  sox_sample_t *mwp=obuf+(*osamp-=*osamp%frame->channels); \
  id##_t x; \
  double y; \
  SOX_SAMPLE_LOCALS; \
 \
  if (0==ar->count) \
    priv->sample=0; \
 \
  sample=priv->sample; \
 \
  while (wp<mwp) { \
    while (sample==frame->nb_samples) { \
      ffsox_ar_next(ar); \
 \
      if (!ffsox_ar_valid(ar)) \
        goto done; \
 \
      sample=priv->sample=0; \
    } \
 \
    memcpy(&x,((id##_t *)frame->data[0])+sample,sizeof x); \
    ++sample; \
 \
    if (1.0==q1) \
      *wp++=SOX_##ID##_TO_SAMPLE(x,effp->clips); \
    else { \
      y=q2*x; \
      *wp++=SOX_FLOAT_64BIT_TO_SAMPLE(y,effp->clips); \
    } \
  } \
done: \
  \
  priv->sample=sample; \
 \
  return 0<(*osamp=wp-obuf)?SOX_SUCCESS:SOX_EOF; \
}
#endif // ]

PRIV_DRAIN_I(s16,SIGNED_16BIT)
PRIV_DRAIN_I(s32,SIGNED_32BIT)
PRIV_DRAIN_I(f32,FLOAT_32BIT)
PRIV_DRAIN_I(f64,FLOAT_64BIT)
#endif // ]

///////////////////////////////////////////////////////////////////////////////
#if defined (LIBFFSOX_BS1770) // [
#define PRIV_DRAIN_P(id1,id2,ID) \
static int priv_drain_##id1##p(sox_effect_t *effp, int nch, \
    sox_sample_t *obuf, size_t *osamp) \
{ \
  priv_t *priv=effp->priv; \
  ffsox_ar_t *ar=priv_ar(priv); \
  double q1=ar->q; \
  double q2=priv_scale_##id1(q1); \
  uint8_t **rp=priv->p.rp; \
  int n,i; \
  sox_sample_t *wp=obuf; \
  sox_sample_t *mwp=obuf+(*osamp-=*osamp%nch); \
  int ch; \
  id1##_t x; \
  double y; \
  bs1770_sample_##id2##_t sample; \
  SOX_SAMPLE_LOCALS; \
 \
  if (0==ar->count) { \
    priv->p.n=0; \
    priv->p.i=0; \
  } \
 \
  n=priv->p.n; \
  i=priv->p.i; \
 \
  while (wp<mwp) { \
    while (i==n) { \
      ffsox_ar_next(ar); \
 \
      if (!ffsox_ar_valid(ar)) \
        goto done; \
 \
      priv_reset_p(priv); \
      rp=priv->p.rp; \
      nch=priv->p.nch; \
      n=priv->p.n; \
      i=priv->p.i; \
    } \
 \
    for (ch=0;ch<nch;++ch) { \
      memcpy(&x,rp[ch]+i*sizeof x,sizeof x); \
      sample[ch]=x; \
 \
      if (1.0==q1) \
        *wp++=SOX_##ID##_TO_SAMPLE(x,effp->clips); \
      else { \
        y=q2*x; \
        *wp++=SOX_FLOAT_64BIT_TO_SAMPLE(y,effp->clips); \
      } \
    } \
 \
    if (NULL!=ar->lufs) \
      ar->lufs->vmt->add_sample_##id2(ar->lufs,sample); \
 \
    ++i; \
  } \
done: \
 \
  priv->p.i=i; \
 \
  return 0<(*osamp=wp-obuf)?SOX_SUCCESS:SOX_EOF; \
}

PRIV_DRAIN_P(s16,i16,SIGNED_16BIT)
PRIV_DRAIN_P(s32,i32,SIGNED_32BIT)
PRIV_DRAIN_P(f32,f32,FLOAT_32BIT)
PRIV_DRAIN_P(f64,f64,FLOAT_64BIT)
#else // ] [
#if 0 // [
#define PRIV_DRAIN_P(id,ID) \
static int priv_drain_##id##p(sox_effect_t *effp, int nch, \
    sox_sample_t *obuf, size_t *osamp) \
{ \
  priv_t *priv=effp->priv; \
  ffsox_ar_t *ar=priv_ar(priv); \
  double q1=ar->q; \
  double q2=priv_scale_##id(q1); \
  uint8_t **rp=priv->p.rp; \
  int n,i; \
  sox_sample_t *wp=obuf; \
  sox_sample_t *mwp=obuf+(*osamp-=*osamp%nch); \
  int ch; \
  id##_t x; \
  double y; \
  SOX_SAMPLE_LOCALS; \
 \
  if (0==ar->count) { \
    priv->p.n=0; \
    priv->p.i=0; \
  } \
 \
  n=priv->p.n; \
  i=priv->p.i; \
 \
  while (wp<mwp) { \
    while (i==n) { \
      ffsox_ar_next(ar); \
 \
      if (!ffsox_ar_valid(ar)) \
        goto done; \
 \
      priv_reset_p(priv); \
      rp=priv->p.rp; \
      nch=priv->p.nch; \
      n=priv->p.n; \
      i=priv->p.i; \
    } \
 \
    for (ch=0;ch<nch;++ch) { \
      memcpy(&x,rp[ch]+i*sizeof x,sizeof x); \
 \
      if (1.0==q1) \
        *wp++=SOX_##ID##_TO_SAMPLE(x,effp->clips); \
      else { \
        y=q2*x; \
        *wp++=SOX_FLOAT_64BIT_TO_SAMPLE(y,effp->clips); \
      } \
    } \
 \
    ++i; \
  } \
done: \
 \
  priv->p.i=i; \
 \
  return 0<(*osamp=wp-obuf)?SOX_SUCCESS:SOX_EOF; \
}
#else // ] [
#define _STR(id) #id
#define STR(id) _STR(id)

#if defined (PRIV_UNION) // [
#define PRIV_DRAIN_P(id,ID) \
static int priv_drain_##id##p(sox_effect_t *effp, int nch, \
    sox_sample_t *obuf, size_t *osamp) \
{ \
  priv_t *priv=effp->priv; \
  ffsox_ar_t *ar=priv_ar(priv); \
  f64_t q1=ar->q; \
  f64_t q2=priv_scale_##id(q1); \
  uint8_t **rp=priv->p.rp; \
  int n,i; \
  sox_sample_t *wp=obuf; \
  sox_sample_t *mwp=obuf+(*osamp-=*osamp%nch); \
  int ch; \
  id##_t x; \
  f64_t y; \
  SOX_SAMPLE_LOCALS; \
 \
  DWRITELN(STR(id##_t)); \
 \
  if (0==ar->count) { \
    priv->p.n=0; \
    priv->p.i=0; \
  } \
 \
  n=priv->p.n; \
  i=priv->p.i; \
 \
  while (wp<mwp) { \
    while (i==n) { \
      ffsox_ar_next(ar); \
 \
      if (!ffsox_ar_valid(ar)) \
        goto done; \
 \
      priv_reset_p(priv); \
      rp=priv->p.rp; \
      nch=priv->p.nch; \
      n=priv->p.n; \
      i=priv->p.i; \
    } \
 \
    for (ch=0;ch<nch;++ch) { \
      memcpy(&x,rp[ch]+i*sizeof x,sizeof x); \
 \
      if (1.0==q1) \
        *wp++=SOX_##ID##_TO_SAMPLE(x,effp->clips); \
      else { \
        y=q2*x; \
DVWRITELN("y=%lf",y); \
        *wp++=SOX_FLOAT_64BIT_TO_SAMPLE(y,effp->clips); \
      } \
    } \
 \
    ++i; \
  } \
done: \
 \
  priv->p.i=i; \
 \
  return 0<(*osamp=wp-obuf)?SOX_SUCCESS:SOX_EOF; \
}
#else // ] [
#define PRIV_DRAIN_P(id,ID) \
static int priv_drain_##id##p(sox_effect_t *effp, sox_sample_t *obuf, \
    size_t *osamp) \
{ \
  priv_t *priv=effp->priv; \
  AVFrame *frame=priv->frame; \
  ffsox_ar_t *ar=priv_ar(priv); \
  f64_t q1=ar->q; \
  f64_t q2=priv_scale_##id(q1); \
  /*uint8_t **rp=priv->p.rp;*/ \
  int sample,channel; \
  /*int n,i;*/ \
  sox_sample_t *wp=obuf; \
  sox_sample_t *mwp=obuf+(*osamp-=*osamp%frame->channels); \
  /*int ch;*/ \
  id##_t x; \
  f64_t y; \
  SOX_SAMPLE_LOCALS; \
 \
  DWRITELN(STR(id##_t)); \
 \
  if (0==ar->count) { \
    /*priv->p.n=0; \
    priv->p.i=0*/; \
    priv->sample=0; \
    priv->channel=0; \
  } \
 \
  /*n=priv->p.n; \
  i=priv->p.i;*/ \
  sample=priv->sample; \
  channel=priv->channel; \
 \
  while (wp<mwp) { \
    while (sample==frame->linesize[0]) { \
      ffsox_ar_next(ar); \
 \
      if (!ffsox_ar_valid(ar)) \
        goto done; \
 \
      /*priv_reset_p(priv); \
      rp=priv->p.rp; \
      nch=priv->p.nch; \
      n=priv->p.n; \
      i=priv->p.i;*/ \
      priv->sample=sample=0; \
      priv->channel=channel=0; \
    } \
 \
    for (channel=0;channel<frame->channels;++channel) { \
      memcpy(&x,((id##_t *)frame->data[channel])+sample,sizeof x); \
 \
      if (1.0==q1) \
        *wp++=SOX_##ID##_TO_SAMPLE(x,effp->clips); \
      else { \
        y=q2*x; \
DVWRITELN("y=%lf",y); \
        *wp++=SOX_FLOAT_64BIT_TO_SAMPLE(y,effp->clips); \
      } \
    } \
 \
    ++sample; \
  } \
done: \
 \
  priv->sample=sample; \
 \
  return 0<(*osamp=wp-obuf)?SOX_SUCCESS:SOX_EOF; \
}
#endif // ]
#endif // ]

PRIV_DRAIN_P(s16,SIGNED_16BIT)
PRIV_DRAIN_P(s32,SIGNED_32BIT)
PRIV_DRAIN_P(f32,FLOAT_32BIT)
PRIV_DRAIN_P(f64,FLOAT_64BIT)
#endif // ]

///////////////////////////////////////////////////////////////////////////////
static int priv_drain(sox_effect_t *effp, sox_sample_t *obuf, size_t *osamp)
{
  priv_t *priv=effp->priv;
#if defined (FFSOX_DEPRECATED_STREAM_CODEC) // [
  AVCodecParameters *codecpar=ffsox_ar_stream(priv_ar(priv))->codecpar;
  int sample_fmt=codecpar->format;
#else // ] [
  AVCodecContext *codecpar=ffsox_ar_stream(priv_ar(priv))->codec;
  int sample_fmt=codecpar->sample_fmt;
#endif // ]
#if defined (PRIV_UNION) // [
  int nch=codecpar->channels;

DVWRITELN("sample_fmt=%s nch=%d",av_get_sample_fmt_name(sample_fmt),nch);

  switch (sample_fmt) {
  // interleaved.
  case AV_SAMPLE_FMT_S16:
    return priv_drain_s16i(effp,nch,obuf,osamp);
  case AV_SAMPLE_FMT_S32:
    return priv_drain_s32i(effp,nch,obuf,osamp);
  case AV_SAMPLE_FMT_FLT:
    return priv_drain_f32i(effp,nch,obuf,osamp);
  case AV_SAMPLE_FMT_DBL:
    return priv_drain_f64i(effp,nch,obuf,osamp);
  // planar.
  case AV_SAMPLE_FMT_S16P:
    return priv_drain_s16p(effp,nch,obuf,osamp);
  case AV_SAMPLE_FMT_S32P:
    return priv_drain_s32p(effp,nch,obuf,osamp);
  case AV_SAMPLE_FMT_FLTP:
    return priv_drain_f32p(effp,nch,obuf,osamp);
  case AV_SAMPLE_FMT_DBLP:
    return priv_drain_f64p(effp,nch,obuf,osamp);
  default:
    return SOX_EOF;
  }
#else // ] [
DVWRITELN("sample_fmt=%s",av_get_sample_fmt_name(sample_fmt));

  switch (sample_fmt) {
  // interleaved.
  case AV_SAMPLE_FMT_S16:
    return priv_drain_s16i(effp,obuf,osamp);
  case AV_SAMPLE_FMT_S32:
    return priv_drain_s32i(effp,obuf,osamp);
  case AV_SAMPLE_FMT_FLT:
    return priv_drain_f32i(effp,obuf,osamp);
  case AV_SAMPLE_FMT_DBL:
    return priv_drain_f64i(effp,obuf,osamp);
  // planar.
  case AV_SAMPLE_FMT_S16P:
    return priv_drain_s16p(effp,obuf,osamp);
  case AV_SAMPLE_FMT_S32P:
    return priv_drain_s32p(effp,obuf,osamp);
  case AV_SAMPLE_FMT_FLTP:
    return priv_drain_f32p(effp,obuf,osamp);
  case AV_SAMPLE_FMT_DBLP:
    return priv_drain_f64p(effp,obuf,osamp);
  default:
    return SOX_EOF;
  }
#endif // ]
}

///////////////////////////////////////////////////////////////////////////////
sox_effect_handler_t const *ffsox_sox_reader()
{
  static sox_effect_handler_t handler={
    // Effect name
    .name="ffsox_sox_reader",
    // Short explanation of parameters accepted by effect
    .usage=NULL,
    // Combination of SOX_EFF_* flags
    .flags=SOX_EFF_MCHAN,
    // Called to parse command-line arguments (called once per effect).
    .getopts=priv_getopts,
    // Called to initialize effect (called once per flow).
    .start=NULL,
    // Called to process samples.
    .flow=NULL,
    // Called to finish getting output after input is complete.
    .drain=priv_drain,
    // Called to shut down effect (called once per flow).
    .stop=NULL,
    // Called to shut down effect (called once per effect).
#if defined (FFSOX_LOCAL_AR) // {
    .kill=priv_kill,
#else // } {
    .kill=NULL,
#endif // }
    // Size of private data SoX should pre-allocate for effect
    .priv_size=sizeof (priv_t)
  };

DWRITELN("==========================================================");
  return &handler;
}
